#include "zink_instance.h"
#include "zink_screen.h"

VkInstance
zink_create_instance(struct zink_instance_info *instance_info)
{
   /* reserve one slot for MoltenVK */
   const char *layers[3] = { 0 };
   uint32_t num_layers = 0;
   
   const char *extensions[4] = { 0 };
   uint32_t num_extensions = 0;

   bool have_EXT_debug_utils = false;
   bool have_KHR_get_physical_device_properties2 = false;
   bool have_MVK_moltenvk = false;

   bool have_layer_KHRONOS_validation = false;
   bool have_layer_LUNARG_standard_validation = false;

#if defined(MVK_VERSION)
   bool have_moltenvk_layer = false;
#endif

   // Build up the extensions from the reported ones but only for the unnamed layer
   uint32_t extension_count = 0;
   if (vkEnumerateInstanceExtensionProperties(NULL, &extension_count, NULL) == VK_SUCCESS) {
       VkExtensionProperties *extension_props = malloc(extension_count * sizeof(VkExtensionProperties));
       if (extension_props) {
           if (vkEnumerateInstanceExtensionProperties(NULL, &extension_count, extension_props) == VK_SUCCESS) {
              for (uint32_t i = 0; i < extension_count; i++) {
                if (!strcmp(extension_props[i].extensionName, "VK_EXT_debug_utils")) {
                    have_EXT_debug_utils = true;
                    extensions[num_extensions++] = "VK_EXT_debug_utils";
                }
                if (!strcmp(extension_props[i].extensionName, "VK_KHR_get_physical_device_properties2")) {
                    have_KHR_get_physical_device_properties2 = true;
                    extensions[num_extensions++] = "VK_KHR_get_physical_device_properties2";
                }
                if (!strcmp(extension_props[i].extensionName, "VK_MVK_moltenvk")) {
                    have_MVK_moltenvk = true;
                    extensions[num_extensions++] = "VK_MVK_moltenvk";
                }
              }
           }
       free(extension_props);
       }
   }

   // Clear have_EXT_debug_utils if we do not want debug info
   if (!(zink_debug & ZINK_DEBUG_VALIDATION)) {
      have_EXT_debug_utils = false;
   }

    // Build up the layers from the reported ones
    uint32_t layer_count = 0;

    if (vkEnumerateInstanceLayerProperties(&layer_count, NULL) == VK_SUCCESS) {
        VkLayerProperties *layer_props = malloc(layer_count * sizeof(VkLayerProperties));
        if (layer_props) {
            if (vkEnumerateInstanceLayerProperties(&layer_count, layer_props) == VK_SUCCESS) {
               for (uint32_t i = 0; i < layer_count; i++) {
                  if (!strcmp(layer_props[i].layerName, "VK_LAYER_KHRONOS_validation")) {
                     have_layer_KHRONOS_validation = true;
                  }
                  if (!strcmp(layer_props[i].layerName, "VK_LAYER_LUNARG_standard_validation")) {
                     have_layer_LUNARG_standard_validation = true;
                  }
#if defined(MVK_VERSION)
                  if (!strcmp(layer_props[i].layerName, "MoltenVK")) {
                     have_moltenvk_layer = true;
                     layers[num_layers++] = "MoltenVK";
                  }
#endif
               }
            }
        free(layer_props);
        }
    }

   instance_info->have_EXT_debug_utils = have_EXT_debug_utils;
   instance_info->have_KHR_get_physical_device_properties2 = have_KHR_get_physical_device_properties2;
   instance_info->have_MVK_moltenvk = have_MVK_moltenvk;

   if (have_layer_KHRONOS_validation && (have_EXT_debug_utils)) {
      layers[num_layers++] = "VK_LAYER_KHRONOS_validation";
      instance_info->have_layer_KHRONOS_validation = true;
   }
   if (have_layer_LUNARG_standard_validation && (have_EXT_debug_utils) && (!have_layer_KHRONOS_validation)) {
      layers[num_layers++] = "VK_LAYER_LUNARG_standard_validation";
      instance_info->have_layer_LUNARG_standard_validation = true;
   }

   VkApplicationInfo ai = {};
   ai.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;

   char proc_name[128];
   if (os_get_process_name(proc_name, ARRAY_SIZE(proc_name)))
      ai.pApplicationName = proc_name;
   else
      ai.pApplicationName = "unknown";

   ai.pEngineName = "mesa zink";
   ai.apiVersion = instance_info->loader_version;

   VkInstanceCreateInfo ici = {};
   ici.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
   ici.pApplicationInfo = &ai;
   ici.ppEnabledExtensionNames = extensions;
   ici.enabledExtensionCount = num_extensions;
   ici.ppEnabledLayerNames = layers;
   ici.enabledLayerCount = num_layers;

   VkInstance instance = VK_NULL_HANDLE;
   VkResult err = vkCreateInstance(&ici, NULL, &instance);
   if (err != VK_SUCCESS)
      return VK_NULL_HANDLE;

   return instance;
}

bool
zink_load_instance_extensions(struct zink_screen *screen)
{
   if (zink_debug & ZINK_DEBUG_VALIDATION) {
      printf("zink: Loader %d.%d.%d \n", VK_VERSION_MAJOR(screen->instance_info.loader_version), VK_VERSION_MINOR(screen->instance_info.loader_version), VK_VERSION_PATCH(screen->instance_info.loader_version));
   }

   if (screen->instance_info.have_KHR_get_physical_device_properties2) {
      if (screen->vk_version < VK_MAKE_VERSION(1,1,0)) {
         GET_PROC_ADDR_INSTANCE_LOCAL(screen->instance, GetPhysicalDeviceFeatures2KHR);
         screen->vk_GetPhysicalDeviceFeatures2 = vk_GetPhysicalDeviceFeatures2KHR;
         if (!screen->vk_GetPhysicalDeviceFeatures2) return false;
         GET_PROC_ADDR_INSTANCE_LOCAL(screen->instance, GetPhysicalDeviceProperties2KHR);
         screen->vk_GetPhysicalDeviceProperties2 = vk_GetPhysicalDeviceProperties2KHR;
         if (!screen->vk_GetPhysicalDeviceProperties2) return false;
         GET_PROC_ADDR_INSTANCE_LOCAL(screen->instance, GetPhysicalDeviceFormatProperties2KHR);
         screen->vk_GetPhysicalDeviceFormatProperties2 = vk_GetPhysicalDeviceFormatProperties2KHR;
         if (!screen->vk_GetPhysicalDeviceFormatProperties2) return false;
         GET_PROC_ADDR_INSTANCE_LOCAL(screen->instance, GetPhysicalDeviceImageFormatProperties2KHR);
         screen->vk_GetPhysicalDeviceImageFormatProperties2 = vk_GetPhysicalDeviceImageFormatProperties2KHR;
         if (!screen->vk_GetPhysicalDeviceImageFormatProperties2) return false;
      } else {
         GET_PROC_ADDR_INSTANCE(GetPhysicalDeviceFeatures2);
         GET_PROC_ADDR_INSTANCE(GetPhysicalDeviceProperties2);
         GET_PROC_ADDR_INSTANCE(GetPhysicalDeviceFormatProperties2);
         GET_PROC_ADDR_INSTANCE(GetPhysicalDeviceImageFormatProperties2);
      }
   }

   return true;
}
